/*-------------------------------------------------------------------------
 Hash function SHA256

 File name   : SHA256.v
 Version     : Version 0.1
 Created     : JUN/09/2007
 Last update : JUN/09/2007
 Desgined by : Akashi Satoh
 
 
 Copyright (C) 2007 AIST
 
 By using this code, you agree to the following terms and conditions.
 
 This code is copyrighted by AIST ("us").
 
 Permission is hereby granted to copy, reproduce, redistribute or
 otherwise use this code as long as: there is no monetary profit gained
 specifically from the use or reproduction of this code, it is not sold,
 rented, traded or otherwise marketed, and this copyright notice is
 included prominently in any copy made.
 
 We shall not be liable for any damages, including without limitation
 direct, indirect, incidental, special or consequential damages arising
 from the use of this code.
 
 When you publish any results arising from the use of this code, we will
 appreciate it if you can cite our webpage
 (http://www.rcis.aist.go.jp/special/SASEBO/index-en.html).
 -------------------------------------------------------------------------*/


//---------------------------------------------
// CSA32 : 32-bit Carry Save Adder (mod 2^32)
//---------------------------------------------
module CSA32 (X, Y, Z, S, C);
  input  [31:0] X;
  input  [31:0] Y;
  input  [31:0] Z;
  output [31:0] S;
  output [31:0] C;

  assign S[31:0] = X[31:0] ^ Y[31:0] ^ Z[31:0];
  assign C[31:1] = ((X[30:0] & Y[30:0]) | (Y[30:0] & Z[30:0])) | (Z[30:0] & X[30:0]);
  assign C[0] = 0;
endmodule


//---------------------------------------------
// CLA4 : 4-bit carry look-ahead adder
//---------------------------------------------

module CLA4 (C, A, B, S, G, P);
  input  C;
  input  [3:0] A;
  input  [3:0] B;
  output [3:0] S;
  output G;
  output P;

  wire g0, g1, g2, g3, p0, p1, p2, p3, c1, c2, c3;

  assign g0 = A[0] & B[0];
  assign g1 = A[1] & B[1];
  assign g2 = A[2] & B[2];
  assign g3 = A[3] & B[3];
  assign p0 = A[0] ^ B[0];
  assign p1 = A[1] ^ B[1];
  assign p2 = A[2] ^ B[2];
  assign p3 = A[3] ^ B[3];
  assign c1 = g0 | (C & p0);
  assign c2 = g1 | (C & p0 & p1) | (g0 & p1);
  assign c3 = g2 | (C & p0 & p1 & p2) | (g0 & p1 & p2) | (g1 & p2);
  assign G  = g3 | (g0 & p1 & p2 & p3) | (g1 & p2 & p3) | (g2 & p3);
  assign P  = p0 & p1 & p2 & p3;

  assign S[0] = p0 ^ C;
  assign S[1] = p1 ^ c1;
  assign S[2] = p2 ^ c2;
  assign S[3] = p3 ^ c3;
endmodule


//---------------------------------------------
// CLA16 : 16-bit carry look-ahead adder
//---------------------------------------------
module CLA16(C, A, B, S, G, P);
  input  C;
  input  [15:0] A;
  input  [15:0] B;
  output [15:0] S;
  output G;
  output P;

  wire g0, g1, g2, g3, p0, p1, p2, p3, c1, c2, c3;

  CLA4 CLA3(c3, A[15:12], B[15:12], S[15:12], g3, p3);
  CLA4 CLA2(c2, A[11: 8], B[11: 8], S[11: 8], g2, p2);
  CLA4 CLA1(c1, A[ 7: 4], B[ 7: 4], S[ 7: 4], g1, p1);
  CLA4 CLA0(C,  A[ 3: 0], B[ 3: 0], S[ 3: 0], g0, p0);

  assign c1 = g0 | (C  & p0);
  assign c2 = g1 | (C  & p0 & p1) | (g0 & p1);
  assign c3 = g2 | (C  & p0 & p1 & p2) | (g0 & p1 & p2) | (g1 & p2);
  assign G  = g3 | (g0 & p1 & p2 & p3) | (g1 & p2 & p3) | (g2 & p3);
  assign P  = p0 & p1 & p2 & p3;
endmodule


//-------------------------------------------------------------------
// CLA32 : 32-bit carry look-ahead adder (no carry in, no carry out)
//-------------------------------------------------------------------
module CLA32(A, B, S);
  input  [32:0] A;
  input  [32:0] B;
  output [32:0] S;

  wire C, g0, g1, p0, p1;

  assign C = 0;

  CLA16 CLA1(g0, A[31:16], B[31:16], S[31:16], g1, p1);
  CLA16 CLA0(C,  A[15: 0], B[15: 0], S[15: 0], g0, p0);
endmodule


//---------------------------------------------
// SIGMA0 : siguma 0 function
//---------------------------------------------
module SIGMA0(X, S);
  input  [31:0] X;
  output [31:0] S;

  assign S = {X[1:0], X[31:2]} ^ {X[12:0], X[31:13]} ^ {X[21:0], X[31:22]};
endmodule


//---------------------------------------------
// SIGMA1 : siguma 1 function
//---------------------------------------------
module SIGMA1(X, S);
  input  [31:0] X;
  output [31:0] S;

  assign S = {X[5:0], X[31:6]} ^ {X[10:0], X[31:11]} ^ {X[24:0], X[31:25]};
endmodule


//---------------------------------------------
// DELTA0 : delta 0 function
//---------------------------------------------
module DELTA0(X, S);
  input  [31:0] X;
  output [31:0] S;

  assign S = {X[6:0], X[31:7]} ^ {X[17:0], X[31:18]} ^ {3'b000, X[31:3]};
endmodule


//---------------------------------------------
// DELTA1 : delta 1 function
//---------------------------------------------
module DELTA1(X, S);
  input  [31:0] X;
  output [31:0] S;

  assign S = {X[16:0], X[31:17]} ^ {X[18:0], X[31:19]} ^ {10'b0000000000, X[31:10]};
endmodule


//---------------------------------------------
// MAJ : Maj|ity function
//---------------------------------------------
module MAJ(X, Y, Z, S);
  input  [31:0] X;
  input  [31:0] Y;
  input  [31:0] Z;
  output [31:0] S;

  assign S = (X & Y) | (X & Z) | (Y & Z);
endmodule


//---------------------------------------------
// CH_F : Ch & f function
//---------------------------------------------
module CH_F(X, Y, Z, S);
  input  [31:0] X;
  input  [31:0] Y;
  input  [31:0] Z;
  output [31:0] S;

  assign S = (X & Y) | (~X & Z);
endmodule


//---------------------------------------------
// MAIN_LOOP : main loop
//---------------------------------------------
module MAIN_LOOP(t, Hinit, Hload, ABCload, Wout, clk, en, Hout);
  input  [5:0]  t;       // integer range 0 to 63;
  input         Hinit;   // Initialize H registers with constants
  input         Hload;   // Enable to load data in H regsters when en = '1'
  input         ABCload; // Enable to load data in a-h regsters when en = '1'
  input  [31:0] Wout;
  input         clk;
  input         en;
  output [31:0] Hout;

  function [31:0] Hconst;
  input [6:0] t;
    case (t)
      0: Hconst = 32'h137e2179;
      1: Hconst = 32'hfb41bd6b;
      2: Hconst = 32'h2b3e6c1f;
      3: Hconst = 32'hade682d1;
      4: Hconst = 32'h5f1d36f1;
      5: Hconst = 32'hfe94f82b;
      6: Hconst = 32'h84caa73b;
      default: Hconst = 32'hf3bcc908;
    endcase
  endfunction

  function [63:0] Kj;
  input [6:0] t;
    case (t)
      0  : Kj = 32'hd728ae22;
      1  : Kj = 32'h23ef65cd;
      2  : Kj = 32'hec4d3b2f;
      3  : Kj = 32'h8189dbbc;
      4  : Kj = 32'hf348b538;
      5  : Kj = 32'hb605d019;
      6  : Kj = 32'haf194f9b;
      7  : Kj = 32'hda6d8118;
      8  : Kj = 32'ha3030242;
      9  : Kj = 32'h45706fbe;
      10 : Kj = 32'h4ee4b28c;
      11 : Kj = 32'hd5ffb4e2;
      12 : Kj = 32'hf27b896f;
      13 : Kj = 32'h3b1696b1;
      14 : Kj = 32'h25c71235;
      15 : Kj = 32'hcf692694;
      16 : Kj = 32'h9ef14ad2;
      17 : Kj = 32'h384f25e3;
      18 : Kj = 32'h8b8cd5b5;
      19 : Kj = 32'h77ac9c65;
      20 : Kj = 32'h592b0275;
      21 : Kj = 32'h6ea6e483;
      22 : Kj = 32'hbd41fbd4;
      23 : Kj = 32'h831153b5;
      24 : Kj = 32'hee66dfab;
      25 : Kj = 32'h2db43210;
      26 : Kj = 32'h98fb213f;
      27 : Kj = 32'hbeef0ee4;
      28 : Kj = 32'h3da88fc2;
      29 : Kj = 32'h930aa725;
      30 : Kj = 32'he003826f;
      31 : Kj = 32'h0a0e6e70;
      32 : Kj = 32'h46d22ffc;
      33 : Kj = 32'h5c26c926;
      34 : Kj = 32'h5ac42aed;
      35 : Kj = 32'h9d95b3df;
      36 : Kj = 32'h8baf63de;
      37 : Kj = 32'h3c77b2a8;
      38 : Kj = 32'h47edaee6;
      39 : Kj = 32'h1482353b;
      40 : Kj = 32'h4cf10364;
      41 : Kj = 32'hbc423001;
      42 : Kj = 32'hd0f89791;
      43 : Kj = 32'h0654be30;
      44 : Kj = 32'hd6ef5218;
      45 : Kj = 32'h5565a910;
      46 : Kj = 32'h5771202a;
      47 : Kj = 32'h32bbd1b8;
      48 : Kj = 32'hb8d2d0c8;
      49 : Kj = 32'h5141ab53;
      50 : Kj = 32'hdf8eeb99;
      51 : Kj = 32'he19b48a8;
      52 : Kj = 32'hc5c95a63;
      53 : Kj = 32'he3418acb;
      54 : Kj = 32'h7763e373;
      55 : Kj = 32'hd6b2b8a3;
      56 : Kj = 32'h5defb2fc;
      57 : Kj = 32'h43172f60;
      58 : Kj = 32'ha1f0ab72;
      59 : Kj = 32'h1a6439ec;
      60 : Kj = 32'h23631e28;
      61 : Kj = 32'hde82bde9;
      62 : Kj = 32'hb2c67915;
      default : Kj = 32'he372532b;
      endcase
  endfunction

  wire [31:0] SIG0, SIG1, MJ, CHF;
  wire [31:0] Kt;
  reg  [31:0] H0, H1, H2, H3, H4, H5, H6, H7;
  reg  [31:0] a, b, c, d, e, f, g, h;
  wire [31:0] H0in, ain, ein, t0, t1, t2, t3, t4;
  wire [31:0] s0, c0, s1, c1, s2, c2, s3, c3, s4, c4;

  SIGMA0 G0(a, SIG0);
  SIGMA1 G1(e, SIG1);
  assign t2 = SIG1;
  MAJ    G2(a, b, c, MJ);
  CH_F   G3(e, f, g, CHF);
  assign t0 = (Hload == 1)? H7: CHF;
  CLA32  G4(h, t0, t1);
  CSA32  G5(t1, t2, Kt, s0, c0);
  CSA32  G6(s0, c0, Wout, s1, c1);
  CSA32  G7(s1, c1, d, s2, c2);
  CLA32  G8(s2, c2, t3);
  CSA32  G9(s1, c1, MJ, s3, c3);
  CSA32  G10(s3, c3, SIG0, s4, c4);
  CLA32  G11(s4, c4, t4);

  assign ain = (Hload == 1)? H0in: t4;
  assign ein = (Hload == 1)? d: t3;

  assign H0in = ((Hload & Hinit) == 1)? Hconst(t): t1;
  assign Hout = H0;
  assign Kt   = Kj(t);  

  always @(posedge clk) begin
      if (en == 1) begin
          if (ABCload == 1) begin
              h <= g;
              g <= f;
              f <= e;
              e <= ein;
              d <= c;
              c <= b;
              b <= a;
              a <= ain;
          end
          if (Hload == 1) begin
              H0 <= H0in;
              H5 <= H4;
              H6 <= H5;
              H7 <= H6;
              H4 <= H3;
              H3 <= H2;
              H2 <= H1;
              H1 <= H0;
          end
      end 
  end
endmodule


//---------------------------------------------
//EY_GEN : key generation loop
//---------------------------------------------
module KEY_GEN(Mi, MSGload, SW, clk, en, Wout);
  input  [31:0] Mi;        // Message input
  input         MSGload;
  input         SW;       // 0:Message in 1:Key ganeration loop
  input         clk;
  input         en;
  output [31:0] Wout;

  reg  [31:0] w [15:0];
  wire [31:0] del0, del1, s0, c0, s1, c1, a, w_in;

  DELTA0 G0(w[1],  del0);
  DELTA1 G1(w[14], del1);
  CSA32  G2(w[0], w[9], del0, s0, c0);
  CSA32  G3(s0, c0, del1, s1, c1);
  CLA32  G4(s1, c1, a);

  assign w_in = (SW == 0)? Mi: a;
  assign Wout = w[15];

  always @(posedge clk) begin
      if ((en & MSGload) == 1) begin
          w[15] <= w_in;
          w[14] <= w[15];
          w[13] <= w[14];
          w[12] <= w[13];
          w[11] <= w[12];
          w[10] <= w[11];
          w[ 9] <= w[10];
          w[ 8] <= w[ 9];
          w[ 7] <= w[ 8];
          w[ 6] <= w[ 7];
          w[ 5] <= w[ 6];
          w[ 4] <= w[ 5];
          w[ 3] <= w[ 4];
          w[ 2] <= w[ 3];
          w[ 1] <= w[ 2];
          w[ 0] <= w[ 1];
      end
  end
endmodule


//---------------------------------------------
// SHA256 : Secure Hash Function SHA-256
//---------------------------------------------
module SHA256(MSG, Mrdy, clk, en, rst, Hout, Hvld);
  input  [31:0] MSG;   // Message input
  input         Mrdy;  // Message input is ready
  input         clk;   // System clock
  input         en;    // Enable
  input         rst;   // Reset
  output [31:0] Hout;  // Hash value output
  output        Hvld;  // Hash value output raedy

  reg       [1:0] state_reg;
  parameter [1:0] WAIT_ST = 0, INIT_ST = 1, MIN_ST = 2, HADD_ST = 3;

  reg  [5:0] t, tx; // integer range 0 to 80;
  reg  Hinit, Hload, ABCload;
  wire MSGload;
  reg  SW, HvldX;
  wire [31:0] Wout, H;

  MAIN_LOOP G0(t, Hinit, Hload, ABCload, Wout, clk, en, H);
  KEY_GEN   G1(MSG, MSGload, SW, clk, en, Wout);

  assign MSGload = ((state_reg == MIN_ST) && (((Mrdy == 1) && (t < 15)) || (t >= 15)))? 1: 0;
  assign Hout = H;
  assign Hvld = HvldX;
  
  always @(posedge clk) begin
     if (rst == 1) begin
         state_reg <= WAIT_ST;
         Hinit   <= 0;
         Hload   <= 0;
         ABCload <= 0;
         SW      <= 0;
         HvldX   <= 0;
         t       <= 0;
         tx      <= 0;
     end
     else if (en == 1) begin
         case (state_reg)

         WAIT_ST: begin
             state_reg <= INIT_ST;
             Hinit     <= 1;
             Hload     <= 1;
             ABCload   <= 1;
         end

         INIT_ST: begin     // H and a-h registers initialize
             if (t == 7) begin
                 Hinit     <= 0;
                 Hload     <= 0;
                 ABCload   <= 0;
                 SW        <= 0;
                 t         <= 0;
                 tx        <= 0;
                 state_reg <= MIN_ST;
             end
             else
                 t <= t + 1;
         end

         MIN_ST: begin       // 80-word message input
             if (tx == 15)
                 SW  <= 1;     // Key Gen Loop
             if (t == 63) begin
                 state_reg <= HADD_ST;
                 Hload     <= 1;
                 ABCload   <= 1;
                 t         <= 0;
                 tx        <= 0;
             end
             else begin
                 if (((Mrdy == 1) && (t < 15)) || (t >= 15)) begin
                     HvldX   <= 0;
                     ABCload <= 1;
                     tx      <= tx + 1;
                     t       <= tx;
                           // One CLK delay is added to t because Message fed throis loaded to W register
                 end
                 else
                     ABCload <= 0;
             end
         end

         HADD_ST: begin
             if (t == 7) begin
                 Hinit     <= 0;
                 Hload     <= 0;
                 ABCload   <= 0;
                 state_reg <= MIN_ST;
                 SW        <= 0;
                 t         <= 0;
             end
             else begin
                 HvldX <= 1;
                 t <= t + 1;
             end
         end

         endcase
      end
  end
endmodule